@@ -13,7 +13,12 @@
 
 #include "vim.h"
 
-#if defined(FEAT_CLIENTSERVER) || defined(PROTO)
+#if defined(FEAT_CLIENTSERVER)
+
+# ifdef FEAT_SOCKETSERVER
+#  include <sys/socket.h>
+#  include "sys/un.h"
+# endif
 
 static void cmdsrv_main(int *argc, char **argv, char_u *serverName_arg, char_u **serverStr);
 static char_u *serverMakeName(char_u *arg, char *cmd);
@@ -37,7 +42,7 @@ server_to_input_buf(char_u *str)
     str = replace_termcodes(str, &ptr, 0, REPTERM_DO_LT, NULL);
     p_cpo = cpo_save;
 
-    if (*ptr != NUL)	// trailing CTRL-V results in nothing
+    if (ptr != NULL && *ptr != NUL)	// trailing CTRL-V results in nothing
     {
 	/*
 	 * Add the string to the input stream.
@@ -70,9 +75,9 @@ eval_client_expr_to_string(char_u *expr)
     funccal_entry_T funccal_entry;
     int		did_save_funccal = FALSE;
 
-#if defined(FEAT_EVAL)
+# if defined(FEAT_EVAL)
     ch_log(NULL, "eval_client_expr_to_string(\"%s\")", expr);
-#endif
+# endif
 
     // Evaluate the expression at the toplevel, don't use variables local to
     // the calling function. Except when in debug mode.
@@ -177,7 +182,7 @@ serverConvert(
 }
 #endif
 
-#if (defined(FEAT_CLIENTSERVER) && !defined(NO_VIM_MAIN)) || defined(PROTO)
+#if defined(FEAT_CLIENTSERVER) && !defined(NO_VIM_MAIN)
 
 /*
  * Common code for the X command server and the Win32 command server.
@@ -191,6 +196,8 @@ static char_u *build_drop_cmd(int filec, char **filev, int tabs, int sendReply);
     void
 exec_on_server(mparm_T *parmp)
 {
+    int made_name = FALSE;
+
     if (parmp->serverName_arg != NULL && *parmp->serverName_arg == NUL)
 	return;
 
@@ -199,6 +206,23 @@ exec_on_server(mparm_T *parmp)
     serverInitMessaging();
 # endif
 
+# ifdef FEAT_SOCKETSERVER
+    // If servername is specified and we are using sockets, always init the
+    // sockt server. We may need to receive replies back to us. If --serverlist
+    // is passed, the socket server will be uninitialized before listing
+    // sockets then initialized after. This is so we don't add our own socket
+    // in the list. This does not happen in serverlist().
+    if ((parmp->serverArg || parmp->serverName_arg != NULL) &&
+	    clientserver_method == CLIENTSERVER_METHOD_SOCKET)
+    {
+	parmp->servername = serverMakeName(parmp->serverName_arg,
+		parmp->argv[0]);
+	if (socket_server_init(parmp->servername) == OK)
+	    TIME_MSG("initialize socket server");
+	made_name = TRUE;
+    }
+# endif
+
     /*
      * When a command server argument was found, execute it.  This may
      * exit Vim when it was successful.  Otherwise it's executed further
@@ -214,8 +238,9 @@ exec_on_server(mparm_T *parmp)
     // If we're still running, get the name to register ourselves.
     // On Win32 can register right now, for X11 need to setup the
     // clipboard first, it's further down.
-    parmp->servername = serverMakeName(parmp->serverName_arg,
-	    parmp->argv[0]);
+    if (!made_name && parmp->servername == NULL)
+	parmp->servername = serverMakeName(parmp->serverName_arg,
+		parmp->argv[0]);
 # ifdef MSWIN
     if (parmp->servername != NULL)
     {
@@ -224,14 +249,13 @@ exec_on_server(mparm_T *parmp)
     }
 # endif
 }
-
 /*
  * Prepare for running as a Vim server.
  */
     void
 prepare_server(mparm_T *parmp)
 {
-# if defined(FEAT_X11)
+# if defined(FEAT_X11) || defined(FEAT_SOCKETSERVER)
     /*
      * Register for remote command execution with :serversend and --remote
      * unless there was a -X or a --servername '' on the command line.
@@ -239,7 +263,13 @@ prepare_server(mparm_T *parmp)
      * or when compiling with autoservername.
      * When running as root --servername is also required.
      */
-    if (X_DISPLAY != NULL && parmp->servername != NULL && (
+
+    if (
+#  ifdef FEAT_X11
+	    X_DISPLAY != NULL &&
+#  endif
+
+	    parmp->servername != NULL && (
 #  if defined(FEAT_AUTOSERVERNAME) || defined(FEAT_GUI)
 		(
 #   if defined(FEAT_AUTOSERVERNAME)
@@ -254,12 +284,26 @@ prepare_server(mparm_T *parmp)
 #  endif
 		parmp->serverName_arg != NULL))
     {
-	(void)serverRegisterName(X_DISPLAY, parmp->servername);
+#  ifdef FEAT_SOCKETSERVER
+	if (clientserver_method == CLIENTSERVER_METHOD_SOCKET)
+	{
+	    if (socket_server_init(parmp->servername) == OK)
+		TIME_MSG("initialize socket server");
+	}
+#  endif
+#  ifdef FEAT_X11
+	if (clientserver_method == CLIENTSERVER_METHOD_X11)
+	{
+	    (void)serverRegisterName(X_DISPLAY, parmp->servername);
+	    TIME_MSG("register x11 server name");
+	}
+#  endif
 	vim_free(parmp->servername);
-	TIME_MSG("register server name");
     }
+#  ifdef FEAT_X11
     else
 	serverDelayedStartName = parmp->servername;
+#  endif
 # endif
 
     /*
@@ -286,22 +330,25 @@ cmdsrv_main(
     char_u	*res;
     int		i;
     char_u	*sname;
-    int		ret;
+    int		ret = -1;
     int		didone = FALSE;
     int		exiterr = 0;
     char	**newArgV = argv + 1;
     int		newArgC = 1,
 		Argc = *argc;
     int		argtype;
-#define ARGTYPE_OTHER		0
-#define ARGTYPE_EDIT		1
-#define ARGTYPE_EDIT_WAIT	2
-#define ARGTYPE_SEND		3
+# define ARGTYPE_OTHER		0
+# define ARGTYPE_EDIT		1
+# define ARGTYPE_EDIT_WAIT	2
+# define ARGTYPE_SEND		3
     int		silent = FALSE;
     int		tabs = FALSE;
-# ifndef FEAT_X11
+# ifdef FEAT_SOCKETSERVER
+    char_u	*receiver;
+# endif
+# ifdef MSWIN
     HWND	srv;
-# else
+# elif defined(FEAT_X11)
     Window	srv;
 
     setup_term_clip();
@@ -384,16 +431,27 @@ cmdsrv_main(
 		}
 		Argc = i;
 	    }
+
+# ifdef FEAT_SOCKETSERVER
+	    if (clientserver_method == CLIENTSERVER_METHOD_SOCKET)
+		ret = socket_server_send(
+			sname, *serverStr, NULL, &receiver,
+			0, -1, silent);
+# endif
 # ifdef FEAT_X11
-	    if (xterm_dpy == NULL)
+	    if (clientserver_method == CLIENTSERVER_METHOD_X11)
 	    {
-		mch_errmsg(_("No display"));
-		ret = -1;
+		if (xterm_dpy == NULL)
+		{
+		    mch_errmsg(_("No display"));
+		    ret = -1;
+		}
+		else
+		    ret = serverSendToVim(xterm_dpy, sname, *serverStr,
+			    NULL, &srv, 0, 0, 0, silent);
 	    }
-	    else
-		ret = serverSendToVim(xterm_dpy, sname, *serverStr,
-						  NULL, &srv, 0, 0, 0, silent);
-# else
+# endif
+# ifdef MSWIN
 	    // Win32 always works?
 	    ret = serverSendToVim(sname, *serverStr, NULL, &srv, 0, 0, silent);
 # endif
@@ -452,14 +510,24 @@ cmdsrv_main(
 		vim_memset(done, 0, numFiles);
 		while (memchr(done, 0, numFiles) != NULL)
 		{
-		    char_u  *p;
+		    char_u  *p = NULL;
 		    int	    j;
 # ifdef MSWIN
 		    p = serverGetReply(srv, NULL, TRUE, TRUE, 0);
 		    if (p == NULL)
 			break;
 # else
-		    if (serverReadReply(xterm_dpy, srv, &p, TRUE, -1) < 0)
+#  ifdef FEAT_SOCKETSERVER
+		    if (clientserver_method == CLIENTSERVER_METHOD_SOCKET
+			    && socket_server_read_reply(receiver, &p, -1) == FAIL)
+			    break;
+#  endif
+#  ifdef FEAT_X11
+		    if (clientserver_method == CLIENTSERVER_METHOD_X11
+			    && serverReadReply(xterm_dpy, srv, &p, TRUE, -1) < 0)
+			    break;
+#  endif
+		    if (p == NULL)
 			break;
 # endif
 		    j = atoi((char *)p);
@@ -490,12 +558,34 @@ cmdsrv_main(
 	    if (serverSendToVim(sname, (char_u *)argv[i + 1],
 						  &res, NULL, 1, 0, FALSE) < 0)
 # else
-	    if (xterm_dpy == NULL)
-		mch_errmsg(_("No display: Send expression failed.\n"));
-	    else if (serverSendToVim(xterm_dpy, sname, (char_u *)argv[i + 1],
-					       &res, NULL, 1, 0, 1, FALSE) < 0)
+#  ifdef FEAT_SOCKETSERVER
+	    if (clientserver_method == CLIENTSERVER_METHOD_SOCKET)
+	    {
+		if (!socket_server_valid())
+		    mch_errmsg(_("Socket server not online:"
+				"Send expression failed"));
+		else if (socket_server_send(sname, (char_u *)argv[i + 1],
+			    &res, NULL, 1, 0, FALSE) < 0)
+		    goto expr_fail;
+	    }
+#  endif
+#  ifdef FEAT_X11
+	    if (clientserver_method == CLIENTSERVER_METHOD_X11)
+	    {
+		if (xterm_dpy == NULL)
+		    mch_errmsg(_("No display: Send expression failed.\n"));
+		else if (serverSendToVim(xterm_dpy, sname,
+			    (char_u *)argv[i + 1], &res,
+			    NULL, 1, 0, 1, FALSE) < 0)
+		    goto expr_fail;
+	    }
+#  endif
+	    if (FALSE)
 # endif
 	    {
+# if !defined(MSWIN)
+expr_fail:
+# endif
 		if (res != NULL && *res != NUL)
 		{
 		    // Output error from remote
@@ -511,8 +601,25 @@ cmdsrv_main(
 	    // Win32 always works?
 	    res = serverGetVimNames();
 # else
-	    if (xterm_dpy != NULL)
+#  ifdef FEAT_SOCKETSERVER
+	    if (clientserver_method == CLIENTSERVER_METHOD_SOCKET)
+	    {
+		int was_init = socket_server_valid();
+
+		// Don't want to add ourselves to the list. So shutdown the
+		// server before listing then startup back again.
+		socket_server_uninit();
+		res = socket_server_list_sockets();
+
+		if (was_init)
+		    socket_server_init(NULL);
+	    }
+#  endif
+#  ifdef FEAT_X11
+	    if (clientserver_method == CLIENTSERVER_METHOD_X11 &&
+		    xterm_dpy != NULL)
 		res = serverGetVimNames(xterm_dpy);
+#  endif
 # endif
 	    if (did_emsg)
 		mch_errmsg("\n");
@@ -541,6 +648,9 @@ cmdsrv_main(
 
     if (didone)
     {
+# ifdef FEAT_SOCKETSERVER
+	socket_server_uninit();
+# endif
 	display_errors();	// display any collected messages
 	exit(exiterr);	// Mission accomplished - get out
     }
@@ -591,11 +701,11 @@ build_drop_cmd(
 	return NULL;
     }
     cdp = vim_strsave_escaped_ext(cwd,
-#ifdef BACKSLASH_IN_FILENAME
+# ifdef BACKSLASH_IN_FILENAME
 	    (char_u *)"",  // rem_backslash() will tell what chars to escape
-#else
+# else
 	    PATH_ESC_CHARS,
-#endif
+# endif
 	    '\\', TRUE);
     vim_free(cwd);
     if (cdp == NULL)
@@ -618,11 +728,11 @@ build_drop_cmd(
 	// do it again in the Vim server.  On MS-Windows only escape
 	// non-wildcard characters.
 	p = vim_strsave_escaped((char_u *)filev[i],
-#ifdef UNIX
+# ifdef UNIX
 		PATH_ESC_CHARS
-#else
+# else
 		(char_u *)" \t%#"
-#endif
+# endif
 		);
 	if (p == NULL)
 	{
@@ -652,13 +762,13 @@ build_drop_cmd(
     //    endif
     //  endif
     ga_concat(&ga, (char_u *)":if !exists('+acd')||!&acd|if haslocaldir()|");
-#ifdef MSWIN
+# ifdef MSWIN
     // in case :set shellslash is set, need to normalize the directory separators
     // '/' is not valid in a filename so replacing '/' by '\\' should be safe
     ga_concat(&ga, (char_u *)"cd -|lcd -|elseif getcwd()->tr('/','\\') ==# '");
-#else
+# else
     ga_concat(&ga, (char_u *)"cd -|lcd -|elseif getcwd() ==# '");
-#endif
+# endif
     ga_concat(&ga, cdp);
     ga_concat(&ga, (char_u *)"'|cd -|endif|endif<CR>");
     vim_free(cdp);
@@ -694,7 +804,24 @@ serverMakeName(char_u *arg, char *cmd)
     char_u *p;
 
     if (arg != NULL && *arg != NUL)
+    {
+# ifdef FEAT_SOCKETSERVER
+	// If we are using a socket server, we want to preserve the original
+	// name if it is a path, else uppercase it if its just a generic name.
+	if (clientserver_method == CLIENTSERVER_METHOD_SOCKET)
+	{
+	    if (arg[0] == '/' || STRNCMP(arg, "./", 2) == 0 ||
+		    STRNCMP(arg, "../", 3) == 0)
+		p = vim_strsave(arg);
+	    else
+		p = vim_strsave_up(arg);
+	}
+	else
+	    p = vim_strsave_up(arg);
+# else
 	p = vim_strsave_up(arg);
+# endif
+    }
     else
     {
 	p = vim_strsave_up(gettail((char_u *)cmd));
@@ -747,7 +874,12 @@ remote_common(typval_T *argvars, typval_T *rettv, int expr)
 # ifdef MSWIN
     HWND	w;
 # else
+#  ifdef FEAT_X11
     Window	w;
+#  endif
+#  ifdef FEAT_SOCKETSERVER
+    char_u	*client = NULL;
+#  endif
 # endif
 
     if (check_restricted() || check_secure())
@@ -768,14 +900,33 @@ remote_common(typval_T *argvars, typval_T *rettv, int expr)
 # ifdef MSWIN
     if (serverSendToVim(server_name, keys, &r, &w, expr, timeout, TRUE) < 0)
 # else
-    if (serverSendToVim(X_DISPLAY, server_name, keys, &r, &w, expr, timeout,
-								  0, TRUE) < 0)
+#  ifdef FEAT_SOCKETSERVER
+    if (clientserver_method == CLIENTSERVER_METHOD_SOCKET)
+	if (socket_server_send(server_name, keys, &r, &client, expr,
+		    timeout * 1000, TRUE) < 0)
+	    goto fail;
+#  endif
+#  ifdef FEAT_X11
+    if (clientserver_method == CLIENTSERVER_METHOD_X11)
+	if (serverSendToVim(X_DISPLAY, server_name, keys, &r, &w, expr, timeout,
+		    0, TRUE) < 0)
+	    goto fail;
+#  endif
 # endif
+# if !defined(MSWIN)
+    if (FALSE)
     {
+fail:
+# else
+    {
+# endif
 	if (r != NULL)
 	{
 	    emsg((char *)r);	// sending worked but evaluation failed
 	    vim_free(r);
+# ifdef FEAT_SOCKETSERVER
+	    vim_free(client);
+# endif
 	}
 	else
 	    semsg(_(e_unable_to_send_to_str), server_name);
@@ -787,23 +938,44 @@ remote_common(typval_T *argvars, typval_T *rettv, int expr)
     if (argvars[2].v_type != VAR_UNKNOWN)
     {
 	dictitem_T	v;
+# if defined(FEAT_SOCKETSERVER)
+	struct sockaddr_un addr;
+	char_u		str[sizeof(addr.sun_path)];
+# else
 	char_u		str[30];
+# endif
 	char_u		*idvar;
 
 	idvar = tv_get_string_chk(&argvars[2]);
 	if (idvar != NULL && *idvar != NUL)
 	{
+	    str[0] = NUL;
+# ifdef MSWIN
 	    sprintf((char *)str, PRINTF_HEX_LONG_U, (long_u)w);
+# else
+#  ifdef FEAT_X11
+	    if (clientserver_method == CLIENTSERVER_METHOD_X11)
+		sprintf((char *)str, PRINTF_HEX_LONG_U, (long_u)w);
+#  endif
+#  ifdef FEAT_SOCKETSERVER
+	    if (clientserver_method == CLIENTSERVER_METHOD_SOCKET)
+		vim_snprintf((char *)str, sizeof(addr.sun_path),
+			"%s", client);
+#  endif
+# endif
 	    v.di_tv.v_type = VAR_STRING;
 	    v.di_tv.vval.v_string = vim_strsave(str);
 	    set_var(idvar, &v.di_tv, FALSE);
 	    vim_free(v.di_tv.vval.v_string);
 	}
     }
+# ifdef FEAT_SOCKETSERVER
+    vim_free(client);
+# endif
 }
 #endif
 
-#if defined(FEAT_EVAL) || defined(PROTO)
+#if defined(FEAT_EVAL)
 /*
  * "remote_expr()" function
  */
@@ -813,7 +985,7 @@ f_remote_expr(typval_T *argvars UNUSED, typval_T *rettv)
     rettv->v_type = VAR_STRING;
     rettv->vval.v_string = NULL;
 
-#ifdef FEAT_CLIENTSERVER
+# ifdef FEAT_CLIENTSERVER
     if (in_vim9script()
 	    && (check_for_string_arg(argvars, 0) == FAIL
 		|| check_for_string_arg(argvars, 1) == FAIL
@@ -823,7 +995,7 @@ f_remote_expr(typval_T *argvars UNUSED, typval_T *rettv)
 	return;
 
     remote_common(argvars, rettv, TRUE);
-#endif
+# endif
 }
 
 /*
@@ -832,11 +1004,11 @@ f_remote_expr(typval_T *argvars UNUSED, typval_T *rettv)
     void
 f_remote_foreground(typval_T *argvars UNUSED, typval_T *rettv UNUSED)
 {
-#ifdef FEAT_CLIENTSERVER
+# ifdef FEAT_CLIENTSERVER
     if (in_vim9script() && check_for_string_arg(argvars, 0) == FAIL)
 	return;
 
-# ifdef MSWIN
+#  ifdef MSWIN
     // On Win32 it's done in this application.
     {
 	char_u	*server_name = tv_get_string_chk(&argvars[0]);
@@ -844,7 +1016,7 @@ f_remote_foreground(typval_T *argvars UNUSED, typval_T *rettv UNUSED)
 	if (server_name != NULL)
 	    serverForeground(server_name);
     }
-# else
+#  else
     // Send a foreground() expression to the server.
     argvars[1].v_type = VAR_STRING;
     argvars[1].vval.v_string = vim_strsave((char_u *)"foreground()");
@@ -853,19 +1025,19 @@ f_remote_foreground(typval_T *argvars UNUSED, typval_T *rettv UNUSED)
     rettv->vval.v_string = NULL;
     remote_common(argvars, rettv, TRUE);
     vim_free(argvars[1].vval.v_string);
+#  endif
 # endif
-#endif
 }
 
     void
 f_remote_peek(typval_T *argvars UNUSED, typval_T *rettv)
 {
-#ifdef FEAT_CLIENTSERVER
+# ifdef FEAT_CLIENTSERVER
     dictitem_T	v;
     char_u	*s = NULL;
-# ifdef MSWIN
+#  ifdef MSWIN
     long_u	n = 0;
-# endif
+#  endif
     char_u	*serverid;
 
     rettv->vval.v_number = -1;
@@ -880,7 +1052,7 @@ f_remote_peek(typval_T *argvars UNUSED, typval_T *rettv)
     serverid = tv_get_string_chk(&argvars[0]);
     if (serverid == NULL)
 	return;		// type error; errmsg already given
-# ifdef MSWIN
+#  ifdef MSWIN
     sscanf((const char *)serverid, SCANF_HEX_LONG_U, &n);
     if (n == 0)
 	rettv->vval.v_number = -1;
@@ -889,13 +1061,22 @@ f_remote_peek(typval_T *argvars UNUSED, typval_T *rettv)
 	s = serverGetReply((HWND)n, FALSE, FALSE, FALSE, 0);
 	rettv->vval.v_number = (s != NULL);
     }
-# else
-    if (check_connection() == FAIL)
-	return;
+#  else
+#   ifdef FEAT_SOCKETSERVER
+    if (clientserver_method == CLIENTSERVER_METHOD_SOCKET)
+	rettv->vval.v_number = socket_server_peek_reply(serverid, &s);
+#   endif
+#   ifdef FEAT_X11
+    if (clientserver_method == CLIENTSERVER_METHOD_X11)
+    {
+	if (check_connection() == FAIL)
+	    return;
 
-    rettv->vval.v_number = serverPeekReply(X_DISPLAY,
-						serverStrToWin(serverid), &s);
-# endif
+	rettv->vval.v_number = serverPeekReply(X_DISPLAY,
+		serverStrToWin(serverid), &s);
+    }
+#   endif
+#  endif
 
     if (argvars[1].v_type != VAR_UNKNOWN && rettv->vval.v_number > 0)
     {
@@ -908,9 +1089,9 @@ f_remote_peek(typval_T *argvars UNUSED, typval_T *rettv)
 	    set_var(retvar, &v.di_tv, FALSE);
 	vim_free(v.di_tv.vval.v_string);
     }
-#else
+# else
     rettv->vval.v_number = -1;
-#endif
+# endif
 }
 
     void
@@ -918,7 +1099,7 @@ f_remote_read(typval_T *argvars UNUSED, typval_T *rettv)
 {
     char_u	*r = NULL;
 
-#ifdef FEAT_CLIENTSERVER
+# ifdef FEAT_CLIENTSERVER
     char_u	*serverid;
 
     if (in_vim9script()
@@ -930,27 +1111,36 @@ f_remote_read(typval_T *argvars UNUSED, typval_T *rettv)
     if (serverid != NULL && !check_restricted() && !check_secure())
     {
 	int timeout = 0;
-# ifdef MSWIN
+#  ifdef MSWIN
 	// The server's HWND is encoded in the 'id' parameter
 	long_u		n = 0;
-# endif
+#  endif
 
 	if (argvars[1].v_type != VAR_UNKNOWN)
 	    timeout = tv_get_number(&argvars[1]);
 
-# ifdef MSWIN
+#  ifdef MSWIN
 	sscanf((char *)serverid, SCANF_HEX_LONG_U, &n);
 	if (n != 0)
 	    r = serverGetReply((HWND)n, FALSE, TRUE, TRUE, timeout);
 	if (r == NULL)
-# else
-	if (check_connection() == FAIL
-		|| serverReadReply(X_DISPLAY, serverStrToWin(serverid),
-						       &r, FALSE, timeout) < 0)
-# endif
 	    emsg(_(e_unable_to_read_server_reply));
+#  else
+#   ifdef FEAT_SOCKETSERVER
+	if (clientserver_method == CLIENTSERVER_METHOD_SOCKET &&
+		socket_server_read_reply(serverid, &r, timeout * 1000) == FAIL)
+	    emsg(_(e_unable_to_read_server_reply));
+#   endif
+#   ifdef FEAT_X11
+	if (clientserver_method == CLIENTSERVER_METHOD_X11 &&
+		(check_connection() == FAIL
+		|| serverReadReply(X_DISPLAY, serverStrToWin(serverid),
+						       &r, FALSE, timeout) < 0))
+	    emsg(_(e_unable_to_read_server_reply));
+#   endif
+#  endif
     }
-#endif
+# endif
     rettv->v_type = VAR_STRING;
     rettv->vval.v_string = r;
 }
@@ -964,7 +1154,7 @@ f_remote_send(typval_T *argvars UNUSED, typval_T *rettv)
     rettv->v_type = VAR_STRING;
     rettv->vval.v_string = NULL;
 
-#ifdef FEAT_CLIENTSERVER
+# ifdef FEAT_CLIENTSERVER
     if (in_vim9script()
 	    && (check_for_string_arg(argvars, 0) == FAIL
 		|| check_for_string_arg(argvars, 1) == FAIL
@@ -972,7 +1162,7 @@ f_remote_send(typval_T *argvars UNUSED, typval_T *rettv)
 	return;
 
     remote_common(argvars, rettv, FALSE);
-#endif
+# endif
 }
 
 /*
@@ -981,7 +1171,7 @@ f_remote_send(typval_T *argvars UNUSED, typval_T *rettv)
     void
 f_remote_startserver(typval_T *argvars UNUSED, typval_T *rettv UNUSED)
 {
-#ifdef FEAT_CLIENTSERVER
+# ifdef FEAT_CLIENTSERVER
     if (check_for_nonempty_string_arg(argvars, 0) == FAIL)
 	return;
 
@@ -992,22 +1182,29 @@ f_remote_startserver(typval_T *argvars UNUSED, typval_T *rettv UNUSED)
     }
 
     char_u *server = tv_get_string_chk(&argvars[0]);
-# ifdef FEAT_X11
-    if (check_connection() == OK)
-	serverRegisterName(X_DISPLAY, server);
-# else
+#  ifdef MSWIN
     serverSetName(server);
-# endif
+#  else
+#   ifdef FEAT_SOCKETSERVER
+    if (clientserver_method == CLIENTSERVER_METHOD_SOCKET)
+	socket_server_init(server);
+#   endif
+#   ifdef FEAT_X11
+    if (clientserver_method == CLIENTSERVER_METHOD_X11 &&
+	    check_connection() == OK)
+	serverRegisterName(X_DISPLAY, server);
+#   endif
+#  endif
 
-#else
+# else
     emsg(_(e_clientserver_feature_not_available));
-#endif
+# endif
 }
 
     void
 f_server2client(typval_T *argvars UNUSED, typval_T *rettv)
 {
-#ifdef FEAT_CLIENTSERVER
+# ifdef FEAT_CLIENTSERVER
     char_u	buf[NUMBUFLEN];
     char_u	*server;
     char_u	*reply;
@@ -1026,20 +1223,37 @@ f_server2client(typval_T *argvars UNUSED, typval_T *rettv)
     if (server == NULL || reply == NULL)
 	return;
 
-# ifdef FEAT_X11
-    if (check_connection() == FAIL)
-	return;
-# endif
+#  ifdef FEAT_SOCKETSERVER
+    if (clientserver_method == CLIENTSERVER_METHOD_SOCKET &&
+	    socket_server_send_reply(server, reply) == FAIL)
+	goto fail;
+#  endif
 
+#  ifdef FEAT_X11
+    if (clientserver_method == CLIENTSERVER_METHOD_X11 &&
+	    check_connection() == FAIL)
+	return;
+
+    if (clientserver_method == CLIENTSERVER_METHOD_X11 &&
+	    serverSendReply(server, reply) < 0)
+#  endif
+#  ifdef MSWIN
     if (serverSendReply(server, reply) < 0)
+#  endif
+#  if defined(FEAT_SOCKETSERVER) && !defined(FEAT_X11) && !defined(MSWIN)
+    if (FALSE)
+#  endif
     {
+#  ifdef FEAT_SOCKETSERVER
+fail:
+#  endif
 	emsg(_(e_unable_to_send_to_client));
 	return;
     }
     rettv->vval.v_number = 0;
-#else
+# else
     rettv->vval.v_number = -1;
-#endif
+# endif
 }
 
     void
@@ -1047,15 +1261,24 @@ f_serverlist(typval_T *argvars UNUSED, typval_T *rettv)
 {
     char_u	*r = NULL;
 
-#ifdef FEAT_CLIENTSERVER
-# ifdef MSWIN
+# ifdef FEAT_CLIENTSERVER
+#  ifdef MSWIN
     r = serverGetVimNames();
-# else
+#  else
+#   ifdef FEAT_SOCKETSERVER
+    if (clientserver_method == CLIENTSERVER_METHOD_SOCKET)
+	r = socket_server_list_sockets();
+#   endif
+#   ifdef FEAT_X11
+    if (clientserver_method == CLIENTSERVER_METHOD_X11)
+    {
     make_connection();
     if (X_DISPLAY != NULL)
 	r = serverGetVimNames(X_DISPLAY);
+    }
+#   endif
+#  endif
 # endif
-#endif
     rettv->v_type = VAR_STRING;
     rettv->vval.v_string = r;
 }
