@@ -76,7 +76,7 @@
 # if RUBY_VERSION >= 20
 // USE_TYPEDDATA is not defined yet. We just check for 2.0.
 #  define rb_check_typeddata		rb_check_typeddata_stub
-#endif
+# endif
 
 # if RUBY_VERSION == 21
 // Ruby 2.1 adds new GC called RGenGC and RARRAY_PTR uses
@@ -179,6 +179,9 @@
 #ifdef HAVE_DUP
 # undef HAVE_DUP
 #endif
+#ifdef HAVE_FSYNC
+# undef HAVE_FSYNC
+#endif
 
 // Avoid redefining TRUE/FALSE in vterm.h.
 #ifdef TRUE
