@@ -13,7 +13,7 @@
 
 #include "vim.h"
 
-#if defined(FEAT_EVAL) || defined(PROTO)
+#if defined(FEAT_EVAL)
 
 
 /*
@@ -572,7 +572,7 @@ parse_generic_func_type_params(
 	return NULL;
     }
 
-    // set the generic parms to VAR_ANY type
+    // set the generic params to VAR_ANY type
     if (ga_grow(&gfatab->gfat_param_types, gfat_sz) == FAIL)
 	return NULL;
 
